import inspect as inspect_
import os
import sys
from dataclasses import dataclass
from inspect import getsourcefile, isfunction

"""
miscellaneous functions that are often used
"""


def default(val, d):
    """If val exists, return it. Otherwise, return d"""
    if val is not None:
        return val
    return d() if isfunction(d) else d


def cycle(dl):
    while True:
        yield from dl


def abspath():
    # https://stackoverflow.com/questions/16771894/python-nameerror-global-name-file-is-not-defined
    # https://docs.python.org/3/library/inspect.html#inspect.FrameInfo
    # return os.path.dirname(inspect.stack()[1][1]) # type: ignore
    # return os.path.dirname(getsourcefile(lambda:0)) # type: ignore
    return os.path.dirname(getsourcefile(inspect_.stack()[1][0]))  # type: ignore


def apply_dict(fn, d, *args, **kwargs):
    return {k: fn(v, *args, **kwargs) for k, v in d.items()}


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


"""
often used modules
"""
# beartype
from beartype import beartype
from beartype.door import die_if_unbearable  # <-- like "assert isinstance(...)"
from beartype.door import is_bearable  # <-------- like "isinstance(...)"
from beartype.vale import Is

enforced_dataclass = beartype(dataclass)
frozen_enforced_dataclass = beartype(dataclass(frozen=True))


from .dict_operations import *
from .itertools import *
from .logging import *
from .math import *
from .serialization import *
from .structures import *



"""
GPU wrappers
"""
import torch

from eztils import default


def device_dtype_decorator(func):
    def wrapper(*args, torch_device=None, **kwargs):
        from eztils.torch import DEVICE, DTYPE

        torch_device = default(torch_device, DEVICE)
        torch_dtype = default(kwargs.get("dtype"), DTYPE)
        kwargs.pop("device", None)
        kwargs.pop("dtype", None)
        return func(*args, **kwargs, device=torch_device, dtype=torch_dtype)

    return wrapper


@device_dtype_decorator
def from_numpy(*args, **kwargs):
    return torch.from_numpy(*args, **kwargs)


@device_dtype_decorator
def tensor(*args, **kwargs):
    return torch.tensor(*args, **kwargs)


@device_dtype_decorator
def randint(*sizes, **kwargs):
    return torch.randint(*sizes, **kwargs)


@device_dtype_decorator
def zeros(*sizes, **kwargs):
    return torch.zeros(*sizes, **kwargs)


@device_dtype_decorator
def empty(*sizes, **kwargs):
    return torch.empty(*sizes, **kwargs)


@device_dtype_decorator
def ones(*sizes, **kwargs):
    return torch.ones(*sizes, **kwargs)


@device_dtype_decorator
def ones_like(*args, **kwargs):
    return torch.ones_like(*args, **kwargs)


@device_dtype_decorator
def randn(*args, **kwargs):
    return torch.randn(*args, **kwargs)


@device_dtype_decorator
def zeros_like(*args, **kwargs):
    return torch.zeros_like(*args, **kwargs)


@device_dtype_decorator
def randperm(*args, **kwargs):
    return torch.randperm(*args, **kwargs)


@device_dtype_decorator
def rand(*args, **kwargs):
    return torch.rand(*args, **kwargs)


@device_dtype_decorator
def tensor(*args, **kwargs):
    return torch.tensor(*args, **kwargs)


@device_dtype_decorator
def normal(*args, **kwargs):
    return torch.normal(*args, **kwargs)



"""
import often used modules
"""
import random
from functools import partial

import numpy as np
import torch
from einops import rearrange as rea
from einops import reduce

try:
    from jaxtyping import Bool, Float, Float16, Float32, Int
except RuntimeError:  # python < 3.9
    pass
from torch import einsum, nn, tensor
from torch.nn import functional as F
from torchvision.utils import save_image

from eztils import load, save

"""
globals
"""
USE_GPU = False
DTYPE = torch.float
GPU_ID = 0
DEVICE = None


def set_gpu_mode(mode, gpu_id=0):
    global USE_GPU
    global DEVICE
    global GPU_ID
    GPU_ID = gpu_id
    USE_GPU = mode
    DEVICE = torch.device("cuda:" + str(gpu_id) if USE_GPU else "cpu")
    if USE_GPU:
        torch.cuda.set_device(gpu_id)


def activation_from_string(string):
    if string == "identity":
        return lambda x: x
    return getattr(nn, string)()


def soft_update_from_to(source, target, tau):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)


"""
helper functions
"""


def seed_everything(seed):
    """
    Set the seed for all the possible random number generators
    for global packages.
    :param seed:
    :return: None
    """
    seed = int(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


save = partial(save, save_fn=torch.save)
torch.load = partial(torch.load, map_location="cpu")
load = partial(load, load_fn=torch.load)

from .distributions import *
from .lightning import *
from .math import *
from .modules import *
from .parameters import *
from .tensor_creators import *
from .to import *



# -----------------------------------------------------------------------------#
# ----------------------------- parameter counting ----------------------------#
# -----------------------------------------------------------------------------#
# https://github.com/allenai/allenact/blob/f00445e4ae8724ccc001b3300af5c56a7f882614/allenact/utils/tensor_utils.py#L1


def param_to_module(param):
    module_name = param[::-1].split(".", maxsplit=1)[-1][::-1]
    return module_name


def report_parameters(model, topk=10):
    def _to_str(num):
        if num >= 1e6:
            return f"{(num/1e6):.2f} M"
        else:
            return f"{(num/1e3):.2f} k"

    counts = {k: p.numel() for k, p in model.named_parameters()}
    n_parameters = sum(counts.values())
    print(f"[ utils/arrays ] Total parameters: {_to_str(n_parameters)}")

    modules = dict(model.named_modules())
    sorted_keys = sorted(counts, key=lambda x: -counts[x])
    max_length = max([len(k) for k in sorted_keys])
    for i in range(topk):
        key = sorted_keys[i]
        count = counts[key]
        module = param_to_module(key)
        print(" " * 8, f"{key:10}: {_to_str(count)} | {modules[module]}")

    remaining_parameters = sum([counts[k] for k in sorted_keys[topk:]])
    print(
        " " * 8,
        f"... and {len(counts)-topk} others accounting for {_to_str(remaining_parameters)} parameters",
    )
    return n_parameters



import os
import pickle
import re
from multiprocessing import shared_memory
from pathlib import Path

import loguru
from varname import ImproperUseError, argname


def remove_shm_from_resource_tracker():
    """
    Monkey patch multiprocessing.resource_tracker so SharedMemory won't be tracked
    More details at: https://bugs.python.org/issue38119
    """
    # pylint: disable=protected-access, import-outside-toplevel
    # Ignore linting errors in this bug workaround hack
    from multiprocessing import resource_tracker

    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return None
        return resource_tracker._resource_tracker.register(name, rtype)

    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return None
        return resource_tracker._resource_tracker.unregister(name, rtype)

    resource_tracker.unregister = fix_unregister
    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


# More details at: https://bugs.python.org/issue38119
remove_shm_from_resource_tracker()


def make_valid_python_var(varStr):
    return re.sub(r"\W|^(?=\d)", "_", varStr)


def save(var, folder=None, memory=False, save_fn=pickle.dump):
    folder = Path(folder or ".").absolute().resolve()
    try:
        varname = make_valid_python_var(argname("var"))
    except ImproperUseError:
        varname = "tensor"

    if not memory:
        fname = str(folder / f"{varname}.pt")
        loguru.logger.info(f"Saving {varname} to {fname}")
        save_fn(var, fname)
    else:
        script = f"""
        from multiprocessing import shared_memory
        handle = shared_memory.SharedMemory(name=\\"{varname}\\", create=False)
        breakpoint()
        """.strip().replace(
            "\n", ";"
        )
        var_bytes = pickle.dumps(var)
        try:
            # see if exists
            s = shared_memory.SharedMemory(name=varname)
            s.close()
            s.unlink()
        except FileNotFoundError:
            pass

        shm = shared_memory.SharedMemory(create=True, size=len(var_bytes), name=varname)

        shm.buf[:] = var_bytes
        loguru.logger.info(f"Saved {varname} to shared memory at /dev/shm/{varname}")

        # so that the shared memory is not deleted when the script exits, as a handle still exits
        os.system(f"""tmux new-session -d -s {varname}""")
        os.system(f"""tmux send-keys -t {varname} "python -c '{script}'" Enter""")


def load(path, memory=False, load_fn=pickle.load, **kwargs):
    if not memory:
        return load_fn(path, **kwargs)
    else:
        shm = shared_memory.SharedMemory(name=path, create=False)
        return pickle.loads(shm.buf)
